from hamonsoft.database._impala.impala import Impala
from hamonsoft.database.db_pool import DBPool


class ImpalaPool(DBPool):
    def __init__(self, max_count=1, **kwargs):
        super().__init__()
        try:
            for _ in range(max_count):
                dbImpl = Impala()
                dbImpl.connect(**kwargs)
                self.queue.put(dbImpl)
                self.allList.append(dbImpl)

        except Exception:
            raise


import queue
import threading
from abc import ABCMeta


class DBPool(metaclass=ABCMeta):
    def __init__(self):
        self.connectLock = threading.Lock()
        self.releaseLock = threading.Lock()

        self.queue = queue.Queue()
        self.allList = []

    def get_session(self, timeout=1):
        try:
            self.connectLock.acquire()
            dbImpl = self.queue.get(timeout=timeout)
            return dbImpl

        except:
            return None
        
        finally:
            self.connectLock.release()
    
    def release(self, dbImpl):
        if dbImpl is not None:
            self.releaseLock.acquire()
            self.queue.put(dbImpl)
            self.releaseLock.release()
    
    def check_queue_size(self):
        return self.queue.qsize()


# -*- coding: utf-8 -*-

import configparser


class ConfigLoader:
    config = configparser.RawConfigParser()

    @staticmethod
    def load(config_file):
        ConfigLoader.config.optionxform = str
        ConfigLoader.config.read(config_file)



import mysql.connector

from hamonsoft.database.db_wrapper import DBWrapper


class MySQL(DBWrapper):
    def __init__(self):
        super().__init__()

    def __del__(self):
        pass

    def connect(self, **kwargs):
        try:
            '''
            self.user = kwargs['user']
            self.password = kwargs['password']
            self.host = kwargs['host']
            self.database = kwargs['database']
            self.port = kwargs['port']
            
            self.connection = mysql.connector.connect(user=self.user, password=self.password,
                                                      host=self.host, database=self.database,
                                                      port=self.port)
            '''
            self.connection = mysql.connector.connect(**kwargs)
        except:
            raise
        
    def reconnect(self):
        try:
            self.connection.reconnect()
            '''
            self.connection = mysql.connector.connect(user=self.user, password=self.password,
                                                      host=self.host, database=self.database,
                                                      port=self.port)
            '''
        except:
            raise

    def check_connection(self):
        try:
            return self.connection is not None and self.connection.is_connected()
        except:
            return False

    def call_proc(self, sp, params):
        try:
            self.cursor = self.connection.cursor()
            return self.cursor.callproc(sp, params)
        except:
            raise

    def execute(self, query, bindValue=None):
        try:
            self.cursor = self.connection.cursor()
            if bindValue is None:
                self.cursor.execute(query, self.bindValue)
            else:
                self.cursor.execute(query, bindValue)

            rowCount = self.cursor.rowcount
            self.cursor.close()

            return rowCount
        except:
            raise
        
    def execute_many(self, query, bindValue):
        try:
            self.cursor = self.connection.cursor()
            self.cursor.executemany(query, bindValue)

            rowCount = self.cursor.rowcount
            self.cursor.close()

            return rowCount
        except:
            raise
    
    def select(self, query, bindValue=None):
        try:
            self.cursor = self.connection.cursor()
            if bindValue is None:
                self.cursor.execute(query, self.bindValue)
            else:
                self.cursor.execute(query, bindValue)

            # columns = [i[0].upper() for i in self.cursor.description ]
            columns = [i[0] for i in self.cursor.description]
            rows = self.cursor.fetchall()

            newRows = [dict(zip(columns, row)) for row in rows]

            self.cursor.close()

            return newRows
        except:
            raise
    
    def select_org(self, query, bindValue=None):
        try:
            self.cursor = self.connection.cursor()
            if bindValue is None:
                self.cursor.execute(query, self.bindValue)
            else:
                self.cursor.execute(query, bindValue)

            rows = self.cursor.fetchall()

            self.cursor.close()

            return rows
        except:
            raise
    
    def add_bind(self, key, value):
        self.bindValue[key] = value
    
    def clear_bind(self):
        self.bindValue.clear()
        
    def close(self):
        self.connection.close()

    def commit(self):
        self.connection.commit()

    def rollback(self):
        self.connection.rollback()



# -*- coding: utf-8 -*-
from hamonsoft.database._impala.impala_pool import ImpalaPool
from hamonsoft.database._mysql.mysql_pool import MySQLPool

from abc import ABCMeta, abstractmethod

class DatabaseType:
    MYSQL = 'mysql'
    IMPALA = 'impala'


class SessionAbstract():
    version = '1_0_0'
    def __init__(self) -> None:
        self.dbPool = None
        
    def set_database(self, _type, max_count=1, **kwargs):
        if _type == DatabaseType.MYSQL:
            self.dbPool = MySQLPool(max_count=max_count, **kwargs)
        elif _type == DatabaseType.IMPALA:
            self.dbPool = ImpalaPool(max_count=max_count, **kwargs)
        else:
            raise ValueError(f'{_type} is not supported')

    @abstractmethod
    def get_session(self, timeout):
        pass

    @abstractmethod
    def release(self, session):
        pass
        
    @abstractmethod
    def close(self, session):
        pass
        
    @abstractmethod
    def check_queue_size(self):
        pass
    
    def check_dbmpl(self):
        return self.dbPool
    


'''
[MariaDb 세션 매니저]
SessionManager(_type=DatabaseType.MYSQL)

[Impala 세션 매니저]
SessionManager(_type=DatabaseType.IMPALA)

<비고>
IMPALA는 RDBMS와 다른 중요한 특성이 있다. 

Connection과 Session 분리:
Hive에서는 Connection과 Session을 별도로 관리한다. 
Connection은 데이터베이스 서버에 대한 연결을 나타내고, Session은 쿼리 실행 및 상태를 추적하는 데 사용된다.
커서는 쿼리를 실행하는 데 사용되며, 각각의 커서는 개별적인 세션을 가진다. 
따라서 커서를 생성할 때마다 새로운 세션을 만들어 해당 세션에서 쿼리를 실행한다.

hs2를 따르면 세션을 관리할 필요가 없다.(내부에서 알아서 한다....)
'''
class SessionManager(SessionAbstract):
    def __init__(self, **parameters):
        super().__init__()
        self.set_database(**parameters)

    def get_session(self, timeout=1):
        # 세션풀에서 사용가능한 세션을 하나 가져온다.
        # 세션매니저는 풀에서 사용가능한 하나의 세션만 가져오고 관리한다.
        # 세션풀에서 가져올 객체가 없는 경우 처리(타임아웃 기간 동안 대기) 로직 추가
        # 세션에 이상이 있는지 확인하고 재접속과 같이 처리한 이후 세션을 리턴시킨다
        session = self.dbPool.get_session(timeout)
        if not session.check_connection():
            session.reconnect()

        return session

    def check_queue_size(self):
        return self.dbPool.check_queue_size()

    def release(self, session):
        if session is not None:
            self.dbPool.release(session)

    def commit(self, session):
        if session is not None:
            session.commit()

    def rollback(self, session):
        if session is not None:
            session.rollback()
            
    def close(self, session):
        self.dbPool.close(session)
        


import logging
import time
from abc import ABC, abstractmethod
from datetime import datetime, timedelta

import mysql.connector


class DatabaseSource(ABC):
    def __init__(self, use_pool: bool = True, pool_size: int = 5, pool_name: str = "default-pool"):
        self._single_connection = None
        self._pool = None

        self._use_pool = use_pool
        self._pool_size = pool_size
        self._pool_name = pool_name
        self._create_con_process()

    @abstractmethod
    def _create_pool_connector(self):
        """
        database의 connection pool을 만드는 로직
        :return:
        """
        pass

    @abstractmethod
    def _create_single_connector(self):
        """
        단일 connection을 만드는 로직
        :return:
        """
        pass

    @abstractmethod
    def get_connection(self):
        """
        최종적으로 사용자가 cursor 메소드만 호출하여 사용할 수 있는 단계로 리턴
        :return:
        """
        pass

    def _create_con_process(self, attempts: int = 3):
        """
        database 생성 시 재시도 로직
        :param attempts:
        :return:
        """
        attempt = 1
        while attempt < attempts + 1:
            try:
                if self._use_pool:
                    self._create_pool_connector()
                else:
                    self._create_single_connector()
                return
            except Exception as e:
                e_message = f"[retry: {attempt}/{attempts}] {str(e)}"
                if attempt == attempts:
                    logging.error(e_message)
                    import sys
                    sys.exit(1)
                logging.error(e_message)
                attempt += 1

    def _basic_exception_handler_with_retry(self, exception: BaseException):
        """
          기본 예외를 처리하고, 커넥션을 재시도
          일정 시간 대기 후에 새로운 데이터베이스 커넥션을 시도
          재시도가 필요한 경우나, 예외 상황에서 자동으로 복구를 시도할 때 사용

          :param exception: 처리할 예외 객체
          :return: 새로운 데이터베이스 커넥션 객체
          """
        import traceback, random
        err_id = random.randint(1000, 9999)

        if isinstance(exception, BaseException):
            logging.warning(f"{err_id}: {str(exception)}...retry... \n {traceback.format_stack()}")
            time.sleep(3)
            return self.get_connection()


class MySql(DatabaseSource):

    def __init__(self, use_pool: bool = False, **kwargs):
        self.kwargs = kwargs
        super().__init__(use_pool=use_pool, pool_size=kwargs.get('pool_size', 5), pool_name=kwargs.get('pool_name', "default-pool"))

    def _create_pool_connector(self):
        self._pool = mysql.connector.pooling.MySQLConnectionPool(**self.kwargs)
        logging.info(f"created connection pool. pool size: {self._pool_size}")

    def _create_single_connector(self):
        self._single_connection = mysql.connector.connect(**self.kwargs)
        logging.info("created a single connection")

    def get_connection(self):
        try:
            if self._use_pool:
                connect = self._pool.get_connection()
            else:
                connect = self._single_connection
            logging.debug(f"get connection: {connect.connection_id}")
            return connect
        except Exception as e:
            return self._basic_exception_handler_with_retry(e)

    @staticmethod
    def self_cursor(func):
        """
        커서를 열고 함수 실행 후 커서와 커넥션을 닫는 데코레이터
        함수에 cursor 객체를 전달
        """

        def wrapper(self, *args, **kwargs):
            self.connection = self.get_connection()
            if self.connection.is_closed():
                raise Exception(f"not connected to database. type: {func.__name__}")
            self.connection_id = self.connection.connection_id
            self.cursor = self.connection.cursor(dictionary=True)

            logging.debug(f"get connection id: {self.connection_id}")
            try:
                result = func(self, *args, **kwargs)
                return result
            except mysql.connector.errors.ProgrammingError as pe:
                self.connection.rollback()
                logging.error(f"error: {pe}")
            except Exception as e:
                self.connection.rollback()
                raise e
            finally:
                self.connection.commit()
                self.cursor.close()
                self.connection.close()
                logging.debug(f"close connection id: {self.connection_id}")

        return wrapper

    @staticmethod
    def self_connection(func):
        """
        커넥션을 열고 함수 실행 후 커넥션을 닫는 데코레이터
        함수에 connection 객체를 전달
        """

        def wrapper(self, *args, **kwargs):
            self.connection = self.get_connection()
            self.connection_id = self.connection.connection_id
            logging.debug(f"get connection id: {self.connection_id}")
            try:
                result = func(self, *args, **kwargs)
                return result
            finally:
                self.connection.close()
                logging.debug(f"close connection id: {self.connection_id}")

        return wrapper

    @staticmethod
    def make_sub_partition_syntax_date_format(start_datetime: datetime, extend_count: int, unit: str = "d") -> str:
        if unit.lower() == "d":
            fmt = "%Y%m%d"
            td = timedelta(days=1)
        elif unit.lower() == "h":
            fmt = "%Y%m%d%H"
            td = timedelta(hours=1)
        else:
            raise ValueError("Invalid unit. 'd' || 'h'")

        partition_list = []
        standard_datetime = start_datetime
        for i in range(extend_count):
            partition_name_str = standard_datetime.strftime(fmt)
            partition_value_str = (standard_datetime + td).strftime(fmt)
            partition_list.append(f"PARTITION p{partition_name_str} VALUES LESS THAN ({partition_value_str})")
            standard_datetime += td

        return ",\n".join(partition_list)



import setuptools

setuptools.setup(
    name="hamonsoft-ai-database",
    version="0.0.18",
    description="Hamonsoft Company Database Python Version Package",
    author="Hamonsoft AI",
    python_requires=">=3.8, <4",
    packages=setuptools.find_packages(),
    install_requires=[
        'mysql-connector-python==9.0.0',
        'impyla==0.19.0'
    ],
    license="MIT",
)



from abc import ABCMeta, abstractmethod


class DBWrapper(metaclass=ABCMeta):
    def __init__(self):
        self.user = None
        self.password = None
        self.host = None
        self.database = None
        self.port = None

        self.connection = None
        self.cursor = None
        self.bindValue = {}

    def __del__(self):
        pass

    @abstractmethod
    def connect(self, *args, **kwargs):
        pass

    @abstractmethod
    def reconnect(self):
        pass

    @abstractmethod
    def check_connection(self):
        pass

    @abstractmethod
    def select_org(self, query, bindValue=None):
        pass

    @abstractmethod
    def select(self, query, bindValue=None):
        pass

    @abstractmethod
    def execute(self, query, bindValue=None):
        pass

    @abstractmethod
    def add_bind(self, key, value):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def clear_bind(self):
        pass

    @abstractmethod
    def commit(self):
        pass

    @abstractmethod
    def rollback(self):
        pass



import impala.dbapi

from hamonsoft.database.db_wrapper import DBWrapper


class Impala(DBWrapper):
    def __init__(self):
        super().__init__()

    def __del__(self):
        pass

    def connect(self, *args, **kwargs):
        try:
            self.host = kwargs['host']
            self.port = kwargs['port']
            self.database = kwargs['database']

            self.connection = impala.dbapi.connect(**kwargs)
        except:
            raise

    def reconnect(self):
        try:
            self.connection.reconnect()
        except:
            raise

    def check_connection(self):
        try:
            result = self.execute('SELECT 1')
            if result:
                return True
            else:
                return False
        except:
            return False

    def execute(self, query, bindValue=None):
        try:
            self.cursor = self.connection.cursor()

            if bindValue is None:
                self.cursor.execute(query, self.bindValue)
            else:
                self.cursor.execute(query, bindValue)

            rowCount = self.cursor.rowcount
            self.cursor.close()

            return rowCount
        except:
            raise

    def execute_many(self, query, bindValue):
        try:
            self.cursor = self.connection.cursor()
            self.cursor.executemany(query, bindValue)

            rowCount = self.cursor.rowcount
            self.cursor.close()

            return rowCount
        except:
            raise

    def select(self, query, bindValue=None):
        try:
            self.cursor = self.connection.cursor()
            if bindValue is None:
                self.cursor.execute(query, self.bindValue)
            else:
                self.cursor.execute(query, bindValue)

            # columns = [i[0].upper() for i in self.cursor.description ]
            columns = [i[0] for i in self.cursor.description]
            rows = self.cursor.fetchall()

            newRows = [dict(zip(columns, row)) for row in rows]

            self.cursor.close()

            return newRows
        except:
            raise

    def select_org(self, query, bindValue=None):
        try:
            self.cursor = self.connection.cursor()
            if bindValue is None:
                self.cursor.execute(query, self.bindValue)
            else:
                self.cursor.execute(query, bindValue)

            rows = self.cursor.fetchall()

            self.cursor.close()

            return rows
        except:
            raise

    def add_bind(self, key, value):
        self.bindValue[key] = value

    def clear_bind(self):
        self.bindValue.clear()

    def close(self):
        self.connection.close()

    def commit(self):
        self.connection.commit()

    def rollback(self):
        self.connection.rollback()
